import subprocess
from pathlib import Path
import click

#  ──────────────────────────────────────────────────────────────────────────

class Database():

    def __init__(self, config, version, initialize=False):
        self.version = version
        self.config = config['database']
        self.path = Path(self.config['path']).expanduser()
        self.host = self.config['host']
        self.cache = Path(self.config['cache']).expanduser()
        self.cache_port = self.config['cache_port']

        if initialize:
            self.init()


    def init(self):
        # make remote store

        # annotated
        command = ['ssh', self.host, 'mkdir -p', self.path / Path('annotated') ]
        subprocess.run(command)

        # supplemental
        command = ['ssh', self.host, 'mkdir -p', self.path / Path('supplemental') ]
        subprocess.run(command)

        # make docker-compose template

        docker_init = r"""version: '3'

services:
    server:
        image: dkllrjr/bibtheque_server:""" + self.version + r"""
        restart: unless-stopped
        volumes:
          - ./files:/bibtheque/files
        ports:
          - "2340:2340"
"""

        docker_init_path = Path('/tmp/docker-compose.yml')
        docker_init_path.open('w').write(docker_init)

        command = ['rsync', '--ignore-existing']
        command += [docker_init_path, self.host + ':' + str(self.path / Path('..'))]
        subprocess.run(command)

        # make local cache
        try:
            # annotated
            cache_path = self.cache / Path('annotated')
            cache_path.mkdir(parents=True)

            # supplemental
            cache_path = self.cache / Path('supplemental')
            cache_path.mkdir(parents=True)

        except:
            click.echo('Cache already exists')


    def write(self, local_file_path, remote_file_path, force=False):

        command = ['rsync']
        
        if force:
            command += ['-I']

        # copying to remote store
        command += [Path(local_file_path).resolve(), self.host + ':' + str(self.path / Path(remote_file_path))]
        subprocess.run(command)

        # moving to local cache
        Path(local_file_path).rename(self.cache / Path(remote_file_path))


    def delete(self, remote_file_path):
        command = ['ssh', self.host, 'rm', self.path / Path(remote_file_path)]
        subprocess.run(command)


    def move(self, old_remote_file_path, new_remote_file_path):
        # moving remote store
        command = ['ssh', self.host, 'mv', self.path / Path(old_remote_file_path), self.path / Path(new_remote_file_path)]
        subprocess.run(command)

        # moving cached version if it exists
        if (self.cache / Path(old_remote_file_path)).exists():
            (self.cache / Path(old_remote_file_path)).rename(self.cache / Path(new_remote_file_path))


    def available(self):
        try:
            command = ['ssh', self.host, 'echo', '"test"', '>', '/tmp/bibtheque_ping']
            subprocess.run(command)
            return True
        except:
            return False


    def clear_cache(self):
        cache_paths = sorted(self.cache.rglob('*.pdf'))
        for path in cache_paths:
            path.unlink()






#  ──────────────────────────────────────────────────────────────────────────
# local imports

import bibtheque.doi as doi
import tests.test_vars as test_vars

# ──────────────────────────────────────────────────────────────────────────
# test functions

def test_is_doi():
    assert doi.is_doi(test_vars.doi_url) == True
    assert doi.is_doi(test_vars.doi_str) == True


def test_doi_to_bib():
    assert doi.doi_to_bib(test_vars.doi_url) == test_vars.doi_bib
    assert doi.doi_to_bib(test_vars.doi_str) == test_vars.doi_bib



# ──────────────────────────────────────────────────────────────────────────
# types

types = ['article',
         'book',
         'booklet',
         'conference',
         'inbook',
         'incollection',
         'inproceedings',
         'manual',
         'mastersthesis',
         'misc',
         'phdthesis',
         'proceedings',
         'techreport',
         'unpublished']


# ──────────────────────────────────────────────────────────────────────────
# fields

required = {'article': ['author',
                        'journal',
                        'title',
                        'type',
                        'volume',
                        'year'],
                   'book': ['author', # removed editor from required
                            'publisher',
                            'title',
                            'type',
                            'year'],
                   'booklet': ['title',
                               'type'],
                   'conference': ['author',
                                  'booktitle',
                                  'title',
                                  'type',
                                  'year'],
                   'inbook': ['author', # removed editor from required
                              'chapter',
                              'pages',
                              'publisher',
                              'title',
                              'type',
                              'year'],
                   'incollection': ['author',
                                    'booktitle',
                                    'publisher',
                                    'title',
                                    'type',
                                    'year'],
                   'inproceedings': ['author',
                                     'booktitle',
                                     'title',
                                     'type',
                                     'year'],
                   'manual': ['title',
                              'type'],
                   'mastersthesis': ['author',
                                     'school',
                                     'title',
                                     'type',
                                     'year'],
                   'misc': ['type'],
                   'phdthesis': ['author',
                                 'school',
                                 'title',
                                 'type',
                                 'year'],
                   'proceedings': ['title',
                                   'type',
                                   'year'],
                   'techreport': ['author',
                                  'institution',
                                  'title',
                                  'type',
                                  'year'],
                   'unpublished': ['author',
                                   'title',
                                   'type',
                                   'note']}


optional = {'article': ['doi',
                        'issn',
                        'key',
                        'month',
                        'note',
                        'number',
                        'pages',
                        'publisher',
                        'url'],
                   'book': ['address',
                            'edition',
                            'editor',
                            'key',
                            'month',
                            'note',
                            'number',
                            'series',
                            'url',
                            'volume'],
                   'booklet': ['address',
                               'author',
                               'howpublished',
                               'key',
                               'month',
                               'note',
                               'year'],
                   'conference': ['address',
                                  'editor',
                                  'key',
                                  'month',
                                  'note',
                                  'number',
                                  'organization',
                                  'pages',
                                  'publisher',
                                  'series',
                                  'volume'],
                   'inbook': ['address',
                              'edition',
                              'key',
                              'month',
                              'note',
                              'number',
                              'series',
                              'volume'],
                   'incollection': ['address',
                                    'chapter',
                                    'edition',
                                    'editor',
                                    'key',
                                    'month',
                                    'note',
                                    'number',
                                    'pages',
                                    'series',
                                    'volume'],
                   'inproceedings': ['address',
                                     'editor',
                                     'key',
                                     'month',
                                     'note',
                                     'number',
                                     'organization',
                                     'pages',
                                     'publisher',
                                     'series',
                                     'volume'],
                   'manual': ['address',
                              'author',
                              'edition',
                              'key',
                              'month',
                              'note',
                              'organization',
                              'year'],
                   'mastersthesis': ['address',
                                     'key',
                                     'month',
                                     'note'],
                   'misc': ['author',
                            'howpublished',
                            'key',
                            'month',
                            'note',
                            'title',
                            'year'],
                   'phdthesis': ['address',
                                 'key',
                                 'month',
                                 'note'],
                   'proceedings': ['address',
                                   'editor',
                                   'key',
                                   'month',
                                   'number',
                                   'publisher',
                                   'series',
                                   'volume'],
                   'techreport': ['address',
                                  'key',
                                  'month',
                                  'note',
                                  'number'],
                   'unpublished': ['key',
                                   'month',
                                   'year']}

#  ──────────────────────────────────────────────────────────────────────────

all_fields = set()

for type in types:
    all_fields.add(type)
    for field in required[type]:
        all_fields.add(field)
    for field in optional[type]:
        all_fields.add(field)



import uuid
import numpy as np
import click

#  ──────────────────────────────────────────────────────────────────────────
# local imports

import bibtheque.doi as doi
import bibtheque.isbn as isbn
import bibtheque.bibtex as bibtex

#  ──────────────────────────────────────────────────────────────────────────

doc_fields = ['bib',
              'file_text',
              'file_path',
              'annotated_path',
              'supplemental_path',
              'tags',
              'notes',
              'synopsis',
              ]

# ──────────────────────────────────────────────────────────────────────────

def determine_standard(standard):
    """
    takes: document standard identifier; DOI or ISBN
    returns: dict or None
    """

    if doi.is_doi(standard):
        doi_bib = doi.doi_to_bib(standard)
        doi_bib_dict = bibtex.bib_to_dict(doi_bib)

        return doi_bib_dict

    elif isbn.is_isbn(standard):
        isbn13 = isbn.get_isbn13(standard)
        isbn_bib = isbn.isbn_to_bib(isbn13)
        isbn_bib_dict = bibtex.bib_to_dict(isbn_bib)

        return isbn_bib_dict

    else:
        raise click.ClickException('Given ID is invalid')


def generate_id(standard):
    """Generate UUID with random salt from the document standard identifier.
    """
    namespace = uuid.UUID('6ba7b812-9dad-11d1-80b4-00c04fd430c8') # namespace for objective identifiers

    salt = str(np.random.random()) # randomized salt

    return uuid.uuid5(namespace, salt + standard).hex


def build_doc(bib_dict, file_text=None, file_path="", annotated_path="", supplemental_path="", tags=[], notes="", synopsis=""):
    """
    Builds doc dict from the bib dict and user supplied things.
    returns: dict
    """

    doc = {}

    doc['bib'] = bib_dict

    # if file_text:
        # doc['file_text'] = file_text

    doc['file_path'] = file_path
    doc['annotated_path'] = annotated_path
    doc['supplemental_path'] = supplemental_path

    # if len(tags) > 0:
    doc['tags'] = tags

    doc['notes'] = notes
    doc['synopsis'] = synopsis

    return doc


def build_bib(bib_dict, bibkey=None):
    """Build the BibTeX string with the given Bibkey.

    Builds a BibTeX string from a BibTex dictionary.

    Generates a Bibkey based off the BibTeX dictionary if no Bibkey is provided.

    Parameters
    ----------
    bib_dict : dict
        BibTeX dictionary of a document.
    bibkey : str
        Unique Bibkey string for the BibTeX string. Defaults to None.

    Returns
    -------
    bib : str
        BibTeX string.
    """

    if bibkey:
        bib_dict['key'] = bibkey

    else:
        bibkey = build_bibkey(bib_dict)

        if not bibkey:
            print("Bibkey not provided and unable to be generated. Please provide a Bibkey.")
            return

        bib_dict['key'] = bibkey
    
    bib = bibtex.build_bib(bib_dict)

    return bib


def lastname(bib_dict):
    """Returns first author's lastname.
    """

    authors = bib_dict['author'].split(' and ')

    first_author = authors[0]
    first_author_lastname = bibtex.author_lastname(first_author)

    return first_author_lastname


def build_bibkey(bib_dict):
    """Builds a unique Bibkey based on the BibTeX dictionary information.

    Fails if the BibTeX dictionary doesn't have an author or year of publication.

    Parameters
    ----------
    bib_dict : dict
        BibTeX dictionary of a document.

    Returns
    -------
    bibkey : str
        Unique Bibkey or None, if generation fails.
    """

    bibkey_list = []
    
    if 'author' in bib_dict:
        bibkey_list.append(lastname(bib_dict))
    
    if 'year' in bib_dict:
        bibkey_list.append(bib_dict['year'])
    
    bibkey = ''.join(bibkey_list)

    if len(bibkey) > 0:
        return bibkey
    else:
        return None

# ──────────────────────────────────────────────────────────────────────────
# print

def et_al(bib_dict):
    """Returns an 'et al.' variant of the authors if there's more than one author.
    """

    if 'author' in bib_dict:
        authors = bib_dict['author'].split(' and ')

        first_author = authors[0]
        first_author_lastname = bibtex.author_lastname(first_author)

        if len(authors) > 1:
            return first_author_lastname + ' et al.'
        else:
            return first_author_lastname

    else:
        return None



from habanero import cn

# ──────────────────────────────────────────────────────────────────────────
# functions

def is_doi(url):
    """Check if a URL is a DOI URL or string.

    Parameters
    ----------
    url : str
        Potential DOI URL or string.

    Returns
    -------
    bool
        True if it is a DOI URL, False if not.
    """

    try:
        cn.content_negotiation(ids=url)
        return True

    except:
        return False


def doi_to_bib(doi):
    """Retrieves the BibTeX citation for a document with a given DOI.

    Parameters
    ----------
    doi : str
        The DOI string of the desired document.

    Returns
    -------
    bib : str
        The BibTeX citation of the given DOI.
    """
    
    try:
        doi_bib = cn.content_negotiation(ids=doi, format='bibtex').strip()

    except:
        raise

    return doi_bib
